from json_cpp import JsonObject, JsonList
from .util import *
import math


class Location(JsonObject):

    def __init__(self,
                 x: float = 0.0,
                 y: float = 0.0):
        self.x = float(x)
        self.y = float(y)

    def __add__(self, o):
        c = Location()
        c.x = self.x + o.x
        c.y = self.y + o.y
        return c

    def __sub__(self, o):
        c = Location()
        c.x = self.x - o.x
        c.y = self.y - o.y
        return c

    def move(self, theta: float, dist: float):
        self.x += math.sin(theta) * dist
        self.y += math.cos(theta) * dist
        return self

    def atan(self, location = None) -> float:
        if location is None:
            return math.atan2(self.x, self.y)
        v = location-self
        return math.atan2(v.x, v.y)

    def mod(self):
        return (self.x ** 2 + self.y ** 2) ** .5

    def dist(self, location=None, segment: tuple = None) -> float:
        if location:
            check_type(location, Location, "incorrect type for location")
            v = location-self
            return (v.x ** 2 + v.y ** 2) ** .5
        elif segment:
            check_type(segment, tuple, "incorrect type for segment")
            line_start, line_end = segment
            check_type(line_start, Location, "incorrect type for line_start")
            check_type(line_end, Location, "incorrect type for line_end")
            normal_length = line_end.dist(line_start)
            distance = ((self.x - line_start.x) * (line_end.y - line_start.y) - (self.y - line_start.y) * (line_end.x - line_start.x)) / normal_length
            return abs(distance)

    def __mul__(self, multiplier):
        if isinstance(multiplier, Location):
            return Location(self.x*multiplier.x, self.y*multiplier.y)
        return Location(self.x * multiplier, self.y * multiplier)

    def __rmul__(self, multiplier):
        if isinstance(multiplier, Location):
            return Location(self.x*multiplier.x, self.y*multiplier.y)
        return Location(self.x * multiplier, self.y * multiplier)

class Scale (Location):
    def __init__(self,
                 x: float = 0.0,
                 y: float = 0.0):
        Location.__init__(self, x=x, y=y)

class Location_list(JsonList):

    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=Location)

    def get_x(self):
        x = []
        for location in self:
            x.append(location.x)
        return x

    def get_y(self):
        y = []
        for location in self:
            y.append(location.y)
        return y


def segments_intersect(segment1: tuple, segment2: tuple):
    segment1_point1, segment1_point2 = segment1
    segment2_point1, segment2_point2 = segment2
    t1 = segment1_point1.atan(segment1_point2)
    t11 = segment1_point1.atan(segment2_point1)
    t12 = segment1_point1.atan(segment2_point2)
    if not angle_between(t1, t11, t12):
        return False
    t2 = segment2_point1.atan(segment2_point2)
    t21 = segment2_point1.atan(segment1_point1)
    t22 = segment2_point1.atan(segment1_point2)
    if not angle_between(t2, t21, t22):
        return False
    return True



import random

from json_cpp import JsonObject, JsonList
from .util import *
from .location import Location
from .coordinates import Coordinates, Coordinates_list

class Cell(JsonObject):
    def __init__(self,
                 cell_id: int = 0,
                 coordinates: Coordinates = None,
                 location: Location = None,
                 occluded: bool = False):
        if coordinates is None:
            coordinates = Coordinates()
        if location is None:
            location = Location()
        check_type(coordinates, Coordinates, "wrong type for coordinates")
        check_type(location, Location, "wrong type for location")
        check_type(cell_id, int, "wrong type for cell_id")
        check_type(occluded, bool, "wrong type for occluded")
        self.id = int(cell_id)
        self.coordinates = coordinates
        self.location = location
        self.occluded = occluded


class Cell_group_builder(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=int)

    @staticmethod
    def get_from_name(world_name: str, name: str, *argv):
        if not type(world_name) is str:
            raise "incorrect type for world_name"
        if not type(name) is str:
            raise "incorrect type for name"
        return Cell_group_builder.parse(json_list=get_resource("cell_group", world_name, name, *argv))


class Cell_group(JsonList):
    def __init__(self, iterable = None,
                 world = None,
                 cell_group_builder: Cell_group_builder = None):
        if cell_group_builder is None:
            cell_group_builder = []
        JsonList.__init__(self, iterable, list_type=Cell)
        if world:
            for cell_id in cell_group_builder:
                self.append(world.cells[cell_id])

    def find(self, location: Location):
        closest = -1
        closest_distance = 0
        for index in range(len(self)):
            dist = location.dist(self[index].location)
            if dist < closest_distance or closest == -1:
                closest = index
                closest_distance = dist
        return closest

    def free_cells(self):
        return self.where("occluded", False)

    def occluded_cells(self):
        return self.where("occluded", True)

    def builder(self):
        cgb = Cell_group_builder()
        for cell in self:
            cgb.append(cell.id)
        return cgb

    def random_cell(self):
        return random.choice(self)


class Cell_map:
    def __init__(self, coordinates_list: Coordinates_list):
        self.coordinates = coordinates_list
        x = self.coordinates.get_x()
        y = self.coordinates.get_x()
        self.base_x = min(x)
        self.base_y = min(y)
        self.max_x = max(x)
        self.max_y = max(y)
        self.size_x = self.max_x - self.base_x + 1
        self.size_y = self.max_y - self.base_y + 1
        self.index = [[-1 for y in range(self.size_y)] for x in range(self.size_x)]
        for i, c in enumerate(coordinates_list):
            self.index[c.x][c.y] = i

    def __getitem__(self, coordinates: Coordinates) -> int:
        try:
            if coordinates.x < self.base_x or coordinates.x > self.max_x:
                return -1
            if coordinates.y < self.base_y or coordinates.y > self.max_y:
                return -1
            return self.index[coordinates.x][coordinates.y]
        except:
            return -1




from json_cpp import JsonObject, JsonList
from .location import *
from .experiment import Episode


class Frame(JsonObject):

    def __init__(self,
                 frame_number: int = 0,
                 time_stamp: float = 0.0,
                 prey_detection: bool = False,
                 predator_detection: bool = False,
                 prey_location: Location = None,
                 predator_location: Location = None,
                 prey_rotation: float = 0.0,
                 predator_rotation: float = 0.0,
                 prey_data: float = "",
                 predator_data: float = ""):
        self.frame_number = frame_number
        self.time_stamp = time_stamp
        self.prey_detection = prey_detection
        self.predator_detection = predator_detection

        if prey_location is None:
            self.prey_location = Location(0, 0)
        else:
            self.prey_location = prey_location

        if predator_location is None:
            self.predator_location = Location(0, 0)
        else:
            self.predator_location = predator_location

        self.prey_rotation = prey_rotation
        self.predator_rotation = predator_rotation
        self.prey_data = prey_data
        self.predator_data = predator_data
        JsonObject.__init__(self)


class Frame_list(JsonList):

    def __init__(self):
        JsonList.__init__(self, list_type=Frame)

    @classmethod
    def from_episode(cls, episode: Episode):
        frame_list = cls()
        last_frame = max(episode.trajectories.get("frame"))
        prey_trajectories = episode.trajectories.get_agent_trajectory("prey")
        predator_trajectories = episode.trajectories.get_agent_trajectory("predator")
        for frame_number in range(last_frame + 1):
            frame = Frame()
            frame.frame_number = frame_number
            prey_step = prey_trajectories.get_step_by_frame(frame_number)
            if prey_step:
                frame.prey_detection = True
                frame.time_stamp = prey_step.time_stamp
                frame.prey_location = prey_step.location
                frame.prey_rotation = prey_step.rotation
                frame.prey_data = prey_step.data
            else:
                frame.prey_detection = False

            predator_step = predator_trajectories.get_step_by_frame(frame_number)
            if predator_step:
                frame.predator_detection = True
                frame.predator_location = predator_step.location
                frame.predator_rotation = predator_step.rotation
                frame.predator_data = predator_step.data
            else:
                frame.predator_detection = False
            frame_list.append(frame)
        return frame_list



import random

from json_cpp import JsonObject, JsonList
from .util import *
from .location import Location, Location_list, Scale
from .coordinates import Coordinates, Coordinates_list
from .shape import Shape, Transformation, Space, Transformation_list
from .cell import Cell, Cell_group_builder, Cell_group, Cell_map


def get_occlusion_count(cell_count, target_entropy):
    for c in range(cell_count):
        p = c/cell_count
        if entropy([p, 1-p])>=target_entropy:
            return c
    return 0


class World_info(JsonObject):

    def __init__(self, world_configuration: str = "", world_implementation: str = "", occlusions: str = ""):
        self.world_configuration = world_configuration
        self.world_implementation = world_implementation
        self.occlusions = occlusions


class World_configuration(JsonObject):

    def __init__(self):
        self.cell_shape = Shape()
        self.cell_coordinates = Coordinates_list()
        self.connection_pattern = Coordinates_list()

    @staticmethod
    def get_from_name(name: str):
        if not type(name) is str:
            raise "incorrect type for name"
        return World_configuration.parse(json_dictionary=get_resource("world_configuration", name))


class World_implementation(JsonObject):

    def __init__(self,
                 cell_locations: Location_list = None,
                 space: Space = None,
                 cell_transformation: Transformation = None):
        if cell_locations is None:
            cell_locations = Location_list()
        if space is None:
            space = Space()
        if cell_transformation is None:
            cell_transformation = Transformation()
        self.cell_locations = cell_locations
        self.space = space
        self.cell_transformation = cell_transformation

    @staticmethod
    def get_from_name(world_configuration_name: str, world_implementation_name: str):
        if not type(world_configuration_name) is str:
            raise "incorrect type for world_configuration_name"
        if not type(world_implementation_name) is str:
            raise "incorrect type for world_implementation_name"
        return World_implementation.parse(json_dictionary=get_resource("world_implementation", world_configuration_name, world_implementation_name))

    @staticmethod
    def create(world_configuration: World_configuration,
               space: Space,
               cell_transformation: Transformation = None,
               relative_locations: Location_list = None,
               relative_locations_transformations: Transformation_list = None,
               center_coordinates: Location_list = None):
        if center_coordinates is None:
            center_coordinates = Coordinates(0, 0)
        if relative_locations is None:
            if relative_locations_transformations is None:
                raise "either relative_locations or relative_locations_transformations must be used"
            relative_locations = World_implementation.create_cell_locations(relative_locations_transformations)
            if len(relative_locations) != len(world_configuration.connection_pattern):
                raise "number of transformations must match the number of connections in the connection pattern"

        if len(world_configuration.connection_pattern) != len(relative_locations):
            raise "the number of locations must match the number of connections"
        wi = World_implementation(space=space, cell_transformation=cell_transformation)
        for x in world_configuration.cell_coordinates:
            wi.cell_locations.append(Location())
        completed = [False for x in world_configuration.cell_coordinates]
        cmap = Cell_map(world_configuration.cell_coordinates)
        wi.cell_locations[cmap[center_coordinates]] = space.center
        set_location(wi.cell_locations, world_configuration.connection_pattern, relative_locations, cmap, center_coordinates, completed)
        return wi

    @staticmethod
    def create_cell_locations(transformations: Transformation_list):
        cell_locations = Location_list()
        for transformation in transformations:
            theta = math.radians(transformation.rotation)
            cell_location = Location(0, 0)
            cell_location.move(theta, transformation.size)
            cell_locations.append(cell_location)
        return cell_locations

    def transform(self,
                   dst_space: Space,
                   cell_transformation: Transformation = None):
        for index, location in enumerate(self.cell_locations):
            new_location = Space.transform_to(location, self.space, dst_space)
            self.cell_locations[index] = new_location
        if cell_transformation is None:
            cell_transformation = self.cell_transformation.copy()
            cell_transformation.size = self.cell_transformation.size * dst_space.transformation.size / self.space.transformation.size
        self.space = dst_space
        self.cell_transformation = cell_transformation

    def scale(self, scale: Scale):
        new_cell_locations = Location_list()
        for cell_location in self.cell_locations:
            new_cell_locations.append(self.space.scale(location=cell_location, scale=scale))
        self.cell_locations = new_cell_locations


class World_statistics(JsonObject):
    def __init__(self,
                 spatial_entropy: float = float(0),
                 spatial_espinometry: float = float(0),
                 spatial_connections: JsonList = None,
                 spatial_connections_derivative: JsonList = None,
                 spatial_centrality: JsonList = None,
                 spatial_centrality_derivative: JsonList = None,
                 visual_entropy: float = float(0),
                 visual_espinometry: float = float(0),
                 visual_connections: JsonList = None,
                 visual_connections_derivative: JsonList = None,
                 visual_centrality: JsonList = None,
                 visual_centrality_derivative: JsonList = None,
                 ITOR_potential: JsonList = None,
                 ITOR_direction: JsonList = None):
        self.spatial_entropy = spatial_entropy
        self.spatial_espinometry = spatial_espinometry
        if spatial_connections:
            self.spatial_connections = spatial_connections
        else:
            self.spatial_connections = JsonList(list_type=int)
        if spatial_connections_derivative:
            self.spatial_connections_derivative = spatial_connections_derivative
        else:
            self.spatial_connections_derivative = JsonList(list_type=int)
        if spatial_centrality:
            self.spatial_centrality = spatial_centrality
        else:
            self.spatial_centrality = JsonList(list_type=float)
        if spatial_centrality_derivative:
            self.spatial_centrality_derivative = spatial_centrality_derivative
        else:
            self.spatial_centrality_derivative = JsonList(list_type=float)

        self.visual_entropy = visual_entropy
        self.visual_espinometry = visual_espinometry
        if visual_connections:
            self.visual_connections = visual_connections
        else:
            self.visual_connections = JsonList(list_type=int)
        if visual_connections_derivative:
            self.visual_connections_derivative = visual_connections_derivative
        else:
            self.visual_connections_derivative = JsonList(list_type=int)
        if visual_centrality:
            self.visual_centrality = visual_centrality
        else:
            self.visual_centrality = JsonList(list_type=float)
        if visual_centrality_derivative:
            self.visual_centrality_derivative = visual_centrality_derivative
        else:
            self.visual_centrality_derivative = JsonList(list_type=float)
        if ITOR_potential:
            self.ITOR_potential = ITOR_potential
        else:
            self.ITOR_potential = JsonList(list_type=float)
        if ITOR_direction:
            self.ITOR_direction = ITOR_direction
        else:
            self.ITOR_direction = JsonList(list_type=Coordinates)

    @staticmethod
    def get_from_parameters_names(
            world_configuration_name: str,
            occlusions_name: str):
        if not type(world_configuration_name) is str:
            raise "incorrect type for world_configuration_name"
        if not type(occlusions_name) is str:
            raise "incorrect type for occlusions_name"
        return World_statistics.parse(json_dictionary=get_resource("world_statistics", world_configuration_name, occlusions_name))


class World:
    def __init__(self, world_configuration: World_configuration):
        self.configuration = world_configuration
        self.implementation = None
        self.cells = Cell_group()
        for cc in world_configuration.cell_coordinates:
            self.cells.append(Cell(cell_id=len(self.cells), coordinates=cc))

    def set_implementation(self, world_implementation):
        self.implementation = world_implementation
        for cell_id, location in enumerate(world_implementation.cell_locations):
            self.cells[cell_id].location = location

    @staticmethod
    def get_from_parameters_names(
            world_configuration_name: str,
            world_implementation_name: str,
            occlusions_name: str = None):
        wc = World_configuration.get_from_name(world_configuration_name)
        o = None
        wi = World_implementation.get_from_name(world_configuration_name, world_implementation_name)
        if occlusions_name:
            o = Cell_group_builder.get_from_name(world_configuration_name, occlusions_name, "occlusions")
        return World.get_from_parameters(wc, wi, o)

    @staticmethod
    def get_from_world_info(world_info: World_info):
        return World.get_from_parameters_names(world_configuration_name=world_info.world_configuration, world_implementation_name=world_info.world_implementation, occlusions_name=world_info.occlusions)


    @staticmethod
    def get_from_parameters(
            world_configuration: World_configuration,
            world_implementation: World_implementation = None,
            occlusions: Cell_group_builder = None):
        wc = world_configuration
        w = World(wc)
        if world_implementation:
            w.set_implementation(world_implementation)

        if occlusions:
            for cell_id in occlusions:
                w.cells[cell_id].occluded = True
        return w

    def create_cell_group(self, cell_group_builder: Cell_group_builder):
        return Cell_group(world=self, cell_group_builder=cell_group_builder)

    def set_occlusions(self, occlusions: Cell_group_builder):
        for cell in self.cells:
            cell.occluded = cell.id in occlusions

    def __str__(self):
        return str(self.cells)

    def get_centrality(self, depth: int = 1) -> JsonList:
        cell_map = Cell_map(self.configuration.cell_coordinates)
        counters = [1 for x in self.cells]
        new_counters = [0 for x in self.cells]
        for i in range(depth):
            for cell_index, cell in enumerate(self.cells):
                if cell.occluded:
                    continue
                for connection in self.configuration.connection_pattern:
                    connection_index = cell_map[cell.coordinates + connection]
                    if connection_index == -1 or self.cells[connection_index].occluded:
                        continue
                    new_counters[cell_index] += counters[connection_index]
            counters = [x for x in new_counters]
        return JsonList(iterable=[c/max(counters) for c in counters], list_type=float)

    def get_centrality_derivative_product(self, depth: int = 1) -> JsonList:
        centrality = self.get_centrality(depth=depth)
        cell_map = Cell_map(self.configuration.cell_coordinates)
        centrality_derivative_product = [0.0 for x in self.cells]
        connection_pairs = self.get_connection_pattern_pairs()
        for cell_index, cell in enumerate(self.cells):
            if cell.occluded:
                continue
            centrality_derivative_product[cell_index] = 1
            for connection, opposite in connection_pairs:
                connection_index = cell_map[cell.coordinates + connection]
                if connection_index == -1 or self.cells[connection_index].occluded:
                    connection_value = 0
                else:
                    connection_value = centrality[connection_index]
                opposite_index = cell_map[cell.coordinates + opposite]
                if opposite_index == -1 or self.cells[opposite_index].occluded:
                    opposite_value = 0
                else:
                    opposite_value = centrality[opposite_index]
                centrality_derivative_product[cell_index] *= abs(connection_value-opposite_value)
        return [d/2 for d in centrality_derivative_product]

    def get_connection_pattern_pairs(self) -> list:
        pairs = []
        for connection in self.configuration.connection_pattern:
            if (-connection, connection) in pairs:
                continue
            pairs.append((connection, -connection))
        return pairs

    def entropy(self, occlusion_count: int, base: float = math.e):
        ent = 0
        cell_count = len(self.cells)
        probabilities = [occlusion_count/cell_count, (cell_count-occlusion_count)/cell_count]
        for p in probabilities:
            if p > 0:
                ent -= p * math.log(p, base)
        return ent / math.log(len(probabilities), base)

    def get_occlusion_count(self, target_entropy: float):
        cell_count = len(self.cells)
        for c in range(cell_count):
            if self.entropy(c) >= target_entropy:
                break
        return c

    def create_occlusions(self, entropy: float, max_walk_length: int = 1):
        if max_walk_length == 0:
            raise "walk length cannot be zero"
        occlusion_count = self.get_occlusion_count(entropy)
        walk = 0
        occlusions = Cell_group_builder()
        m = Cell_map(self.configuration.cell_coordinates)
        for i in range(occlusion_count):
            if walk == 0:
                walk = random.randint(1, max_walk_length)
                curr_cell = self.cells.random_cell()
            else:
                alternatives = []
                for a in self.configuration.connection_pattern:
                    i = m[curr_cell.coordinates + a]
                    if i >= 0:
                        if i not in occlusions:
                            alternatives.append(i)
                if len(alternatives) == 0:
                    curr_cell = self.cells.random_cell()
                    walk = 0
                else:
                    o = random.choice(alternatives)
                    occlusions.append(o)
                    curr_cell = self.cells[o]
                    walk -= 1
        return occlusions


def set_location(
        location_list: Location_list,
        connection_pattern: Coordinates_list,
        relative_locations: Location_list,
        cmap: Cell_map,
        coordinates: Coordinates,
        completed):
    index = cmap[coordinates]
    if completed[index]:
        return
    completed[index] = True
    location = location_list[index]
    for i, c in enumerate(connection_pattern):
        cc = c + coordinates
        cindex = cmap[cc]
        if cindex >= 0:
            location_list[cindex] = location + relative_locations[i]
            set_location(location_list, connection_pattern, relative_locations, cmap, cc, completed)





from .location import Location
from .experiment import Trajectories
from json_cpp import JsonObject, JsonList


class StreamLine(JsonList):
    def __init__(self, trajectory: Trajectories = None, streamline_len: int = 100):
        JsonList.__init__(self, list_type=Location)
        if trajectory is None:
            return
        last_step_location = trajectory[0].location
        total_dist = 0
        for step in trajectory:
            total_dist += last_step_location.dist(step.location)
            last_step_location = step.location
        streamline_step_distance = total_dist / streamline_len
        prev_step_dist = 0
        next_step_dist = 0
        step_index = 0
        prev_step_location = trajectory[0].location
        next_step_location = prev_step_location
        for i in range(streamline_len):
            streamline_step_total_distance = i * streamline_step_distance
            while next_step_dist < streamline_step_total_distance and step_index < len(trajectory):
                prev_step_location = trajectory[step_index].location
                prev_step_dist = next_step_dist
                step_index += 1
                if step_index == len(trajectory):
                    break
                next_step_location = trajectory[step_index].location
                next_step_dist += prev_step_location.dist(next_step_location)
            step_distance = prev_step_location.dist(next_step_location)
            if step_distance > 0:
                prev_ratio = (next_step_dist - streamline_step_total_distance) / step_distance
                next_ratio = (streamline_step_total_distance - prev_step_dist) / step_distance
                next_streamline_step = prev_step_location * prev_ratio + next_step_location * next_ratio
            else:
                next_streamline_step = next_step_location
            self.append(next_streamline_step)
        self.append(trajectory[-1].location)

    def distance(self, streamline) -> float:
        if len(self) != len(streamline):
            raise RuntimeError("StreamLines must have the same length")
        total_distance = 0
        for i, s in enumerate(self):
            total_distance += streamline[i].dist(s)
        return total_distance / len(self)

    @classmethod
    def combine(cls, streamlines):
        combined = StreamLine()
        for s in range(len(streamlines[0])):
            combined_step = Location()
            for sl in streamlines:
                combined_step += sl[s]
            combined_step.x /= len(streamlines)
            combined_step.y /= len(streamlines)
            combined.append(combined_step)
        return combined

    def clone(self):
        r = StreamLine()
        for s in self:
            r.append(Location(s.x, s.y))
        return r


class StreamLineCluster(JsonObject):

    def __iter__(self):
        return self.streamlines.__iter__()

    def __len__(self):
        return len(self.streamlines)

    def __init__(self, streamline: StreamLine = None):
        self.streamlines = JsonList(list_type=StreamLine)
        if streamline:
            self.streamlines.append(streamline)
            self.centroid = streamline
        else:
            self.centroid = StreamLine()
        JsonObject.__init__(self)

    def add_streamline(self, streamline: StreamLine, update_centroid: bool = True) -> None:
        if self.centroid and len(self.centroid) != len(streamline):
            raise RuntimeError("StreamLines must have the same length")
        if update_centroid:
            self.centroid = self.__get_new_centroid__(streamline)
        self.streamlines.append(streamline)

    def __get_new_centroid__(self, streamline: StreamLine) -> StreamLine:
        if not self.centroid:
            return streamline.clone()
        new_centroid = StreamLine()
        streamline_count = len(self.streamlines)
        for i, s in enumerate(self.centroid):
            combined_step = (streamline[i] + s * streamline_count) * (1 / (streamline_count + 1))
            new_centroid.append(combined_step)
        return new_centroid

    def distance(self, streamline: StreamLine, update_centroid=True) -> float:
        if update_centroid:
            new_centroid = self.__get_new_centroid__(streamline)
            return new_centroid.distance(streamline)
        else:
            return self.centroid.distance(streamline)

    def get_distances(self) -> JsonList:
        distances = JsonList(list_type=float)
        for sl in self.streamlines:
            distances.append(sl.distance(self.centroid))
        return distances


class StreamLineClusters(JsonObject):

    def __iter__(self):
        return self.clusters.__iter__()

    def __len__(self):
        return len(self.clusters)

    def __init__(self, min_clusters: int = 1, max_distance: float = .1, streamline_len: int = 100):
        self.min_clusters = min_clusters
        self.max_distance = max_distance
        self.streamline_len = streamline_len
        self.clusters = JsonList(list_type=StreamLineCluster)
        self.unclustered = JsonList(list_type=StreamLine)

    def evaluate_trajectory(self, trajectory: Trajectories) -> StreamLineCluster:
        streamline = StreamLine(trajectory=trajectory, streamline_len=self.streamline_len)
        return self.evaluate_streamline(streamline)

    def get_closest_cluster(self, streamline: StreamLine, update_centroid=True) ->tuple:
        if self.streamline_len + 1 != len(streamline):
            raise RuntimeError("StreamLines must have the same length")
        closest_cluster_distance = 0
        closest_cluster = None
        for c in self.clusters:
            distance = c.distance(streamline, update_centroid)
            if closest_cluster is None or distance < closest_cluster_distance:
                closest_cluster = c
                closest_cluster_distance = distance
        return closest_cluster, closest_cluster_distance

    def evaluate_streamline(self, streamline: StreamLine) -> StreamLineCluster:
        closest_cluster, closest_cluster_distance = self.get_closest_cluster(streamline, True)
        if closest_cluster_distance < self.max_distance:
            return closest_cluster
        else:
            return None

    def add_streamline(self, streamline: StreamLine) -> None:
        candidate_cluster = self.evaluate_streamline(streamline)
        if candidate_cluster is None:
            self.clusters.append(StreamLineCluster(streamline=streamline))
        else:
            candidate_cluster.add_streamline(streamline)

    def filter_clusters(self, threshold: int) -> None:
        self.unclustered = JsonList(list_type=StreamLine)
        while True:
            for i, c in enumerate(self.clusters):
                if len(c) < threshold:
                    break
            else:
                return
            self.unclustered += c
            self.clusters.remove(c)

    def streamline_count(self) -> int:
        counter = 0
        if self.unclustered:
            counter += len(self.unclustered)
        for c in self.clusters:
            counter += len(c.streamlines)
        return counter

    def add_trajectory(self, trajectory: Trajectories) -> None:
        streamline = StreamLine(trajectory=trajectory, streamline_len=self.streamline_len)
        self.add_streamline(streamline)

    def get_average_distance(self):
        distances = self.get_distances()
        if len(distances) == 0:
            return 0
        return sum(distances)/len(distances)

    def get_distances(self) -> JsonList:
        distances = self.get_intra_cluster_distances()
        for usl in self.unclustered:
            c, d = self.get_closest_cluster(usl, False)
            distances.append(d)
        return distances

    def get_intra_cluster_distances(self) -> JsonList:
        distances = JsonList(list_type=float)
        for c in self.clusters:
            distances += c.get_distances()
        return distances

    def get_average_intra_cluster_distance(self):
        distances = self.get_intra_cluster_distances()
        if len(distances) == 0:
            return 0
        return sum(distances)/len(distances)

    def get_cluster_population_standard_deviation(self):
        count = self.streamline_count()
        values = [len(c)/count for c in self.clusters]
        mean = sum(values)/len(values)
        dev = [abs(v-mean) ** 2 for v in values]
        std_dev = (sum(dev)/len(dev)) ** .5
        return std_dev

    def assign_unclustered(self):
        for u in self.unclustered:
            cc, d = self.get_closest_cluster(u)
            cc.add_streamline(u, update_centroid=False)
        self.unclustered.clear()


import networkx as nx
from .cell import Cell_map, Cell_group, Cell, Cell_group_builder
from .world import World


class Graph:

    @staticmethod
    def create_connection_graph(world: World):
        graph = Graph()
        map = Cell_map(world.configuration.cell_coordinates)
        for cell in world.cells:
            src = cell
            src_coord = cell.coordinates
            if not cell.occluded:
                for conn in world.configuration.connection_pattern:
                    dst_coord = src_coord + conn
                    dst_index = map[dst_coord]
                    if dst_index >= 0:
                        dst = world.cells[dst_index]
                        if not dst.occluded:
                            graph.connect(src, dst)
        return graph

    def __init__(self):
        self._connections = []

    def connect(self, src: Cell, dst: Cell, bi=True):
        while src.id >= len(self._connections):
            self._connections.append(Cell_group_builder())
        while dst.id >= len(self._connections):
            self._connections.append(Cell_group_builder())

        if not dst.id in self._connections[src.id]:
            self._connections[src.id].append(dst.id)
        if bi and not src.id in self._connections[dst.id]:
            self._connections[dst.id].append(src.id)

    def is_connected(self, src: Cell, dst: Cell) -> bool:
        if src.id == dst.id:
            return True
        pending = [src.id]
        visited = []
        while pending:
            next_id = pending[0]
            if dst.id in self._connections[next_id]:
                return True
            visited.append(next_id)
            pending = pending[1:]
            pending += [c for c in self._connections[next_id] if not c in visited and not c in pending]
        return False

    def to_nxgraph (self):
        nxgraph = nx.Graph()
        nxgraph.add_nodes_from(list(range(len(self._connections))))
        for src, conns in enumerate(self._connections):
            for dst in conns:
                nxgraph.add_edge(src, dst)
        return nxgraph

    def __getitem__(self, cell: Cell) -> Cell_group_builder:
        return self._connections[cell.id]




from json_cpp import JsonObject, JsonList
from .location import Location, Location_list, segments_intersect, Scale
from .util import *
import math


class Shape(JsonObject):
    def __init__(self, sides=6):
        self.sides = int(sides)


class Transformation(JsonObject):
    def __init__(self, size=0.0, rotation=0.0):
        self.size = float(size)
        self.rotation = float(rotation)


class Transformation_list(JsonList):
    def __init__(self, iterable=None, n=None, size=0.0, rotation=0.0):
        JsonList.__init__(self, iterable, list_type=Transformation)
        if n:
            for a in range(n):
                self.append(Transformation(float(size), float(rotation) + float(a) * (360.0 / float(n))))


class Space(JsonObject):
    def __init__(self,
                 center: Location = None,
                 shape: Shape = None,
                 transformation: Transformation = None):
        self.center = center if center else Location()
        self.shape = shape if shape else Shape()
        self.transformation = transformation if transformation else Transformation()

    @staticmethod
    def transform_to(location: Location, src_space, dst_space):
        check_type(src_space, Space, "incorrect type for src_space")
        check_type(dst_space, Space, "incorrect type for dst_space")
        size_ratio = dst_space.transformation.size / src_space.transformation.size
        rotation = math.radians(dst_space.transformation.rotation - src_space.transformation.rotation)
        dist = src_space.center.dist(location)
        theta = src_space.center.atan(location)
        new_location = Location(dst_space.center.x, dst_space.center.y)
        return new_location.move(theta + rotation, dist * size_ratio)

    def scale(self, location:Location, scale: Scale):
        diff = (location - self.center)
        scaled = diff * scale
        new_location = scaled + self.center
        return new_location

class Polygon:

    def __init__(self,
                 center: Location,
                 sides: int = 0,
                 radius: float = 1.0,
                 rotation: float = 0.0,
                 vertices: Location_list = None):
        if vertices is None:
            vertices = Location_list()
        self.center = center
        self.vertices = vertices
        self.radius = float(radius)
        if sides > 0:
            if len(self.vertices) != 0:
                raise "cannot use sides and vertices together"
            rotation = float(rotation)
            if sides == 0:
                raise "incorrect parameters"
            theta = math.radians(rotation)
            inc = 2.0 * math.pi / sides
            for s in range(sides):
                c = center.copy()
                self.vertices.append(c.move(theta, radius))
                theta += inc
        else:
            if len(self.vertices) == 0:
                raise "must specify either sides or vertices"

    def move(self,
             location: Location = None,
             theta: float = None,
             dist: float = None):
        dif = Location(0, 0)
        if location is None:
            if theta is None or dist is None:
                raise "incorrect parameters"
            dif.move(theta, dist)
        else:
            dif = location - self.center
        self.center = location
        for v in self.vertices:
            v = v + dif

    def contains(self, location: Location):
        dist = self.center.dist(location)
        if dist > self.radius:
            return False
        inner_radius = self.center.dist(segment=(self.vertices[0], self.vertices[1]))
        if dist < inner_radius:
            return True
        for i in range(1, len(self.vertices)):
            if segments_intersect((self.center, location), (self.vertices[i], self.vertices[i-1])):
                return False
        if segments_intersect((self.center, location), (self.vertices[-1], self.vertices[0])):
            return False
        return True

    def is_between(self,
                   src: Location = None,
                   dst: Location = None,
                   theta: float = None,
                   dist: float = None):
        if dst is not None:
            theta = src.atan(dst)
            dist = src.dist(dst)
        else:
            if theta is None or dist is None:
                raise "either dst or theta and dist should be used"
        dist_center = src.dist(self.center)
        theta_center = src.atan(self.center)
        diff_theta_center, direction_center = angle_difference(theta, theta_center)
        if dist < dist_center - self.radius:
            return False
        for v in self.vertices:
            vertex_distance = src.dist(v)
            if vertex_distance < dist:
                theta_vertex = src.atan(v)
                diff_theta_vertex, direction_vertex = angle_difference(theta, theta_vertex);
                if direction_center == -direction_vertex:
                    if diff_theta_center + diff_theta_vertex < math.pi:
                        return True
        return False


class Polygon_list(JsonList):

    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=Polygon)

    @staticmethod
    def get_polygons(centers: Location_list, sides: int, radius: float, rotation: float):
        polygon_list = Polygon_list()
        for center in centers:
            polygon_list.append(Polygon(center,sides=sides, rotation=rotation, radius=radius))
        return polygon_list



from setuptools import setup

setup(name='cellworld',description='Maciver Lab computational biology research package',author='german espinosa',author_email='germanespinosa@gmail.com',packages=['cellworld'],install_requires=['numpy', 'matplotlib', 'json-cpp', 'tcp-messages', 'networkx'],license='MIT',version='0.0.144',zip_safe=False)



from json_cpp import JsonObject, JsonList
from .util import *


class Coordinates(JsonObject):
    def __init__(self,
                 x: int = 0,
                 y: int = 0):
        self.x = int(x)
        self.y = int(y)

    def __add__(self, o):
        c = Coordinates()
        c.x = self.x + o.x
        c.y = self.y + o.y
        return c

    def __sub__(self, o):
        c = Coordinates()
        c.x = self.x - o.x
        c.y = self.y - o.y
        return c

    def __neg__(self):
        c = Coordinates()
        c.x = -self.x
        c.y = -self.y
        return c

    def manhattan(self, o) -> int:
        return abs(o.x - self.x) + abs(o.y - self.y)


class Coordinates_list (JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable, list_type=Coordinates)

    def get_x(self):
        x = []
        for coordinates in self:
            x.append(coordinates.x)
        return x

    def get_y(self):
        y = []
        for coordinates in self:
            y.append(coordinates.y)
        return y

